"""Generic utilities."""

import sys
import os
import shutil


def eprint(*args, **kwargs):
    """Prints to stderr."""
    print(*args, file=sys.stderr, **kwargs)


def move(src_path: str, dst_path: str, overwrite: bool = False, never_overwrite_dir: bool = False):
    """Moves a file or directory.

    There are three built-in functions for moving: os.rename, os.replace and shutil.move.
    The first two do not support moving between different file systems. The last one does, but it
    has possibly confusing semantincs when moving to a directory, and whether or not it overwrites
    an existing file or directory is not configurable and depends on the operating system.

    This function fixes these issues, at the cost of having TOCTOU vulnerabilities.

    If `overwrite==True`, overwrites any existing file at dst_path.
    If, in addition, `never_overwrite_dir==False`, also overwrites any existing directory.
    If `overwrite==False`, changing `never_overwrite_dir` will only affect exception messages.
    """

    if os.path.isfile(dst_path):
        if not overwrite:
            raise FileExistsError(f"File {dst_path} already exists.")
        try:
            shutil.move(src_path, dst_path)
        except FileExistsError:
            os.remove(dst_path)
            shutil.move(src_path, dst_path)
        return

    if os.path.isdir(dst_path):
        if never_overwrite_dir:
            raise FileExistsError(f"Not overwriting directory {dst_path}")
        if not overwrite:
            raise FileExistsError(f"Directory {dst_path} already exists.")
        shutil.rmtree(dst_path)
        shutil.move(src_path, dst_path)
        return

    shutil.move(src_path, dst_path)



"""
Command-line tool and Python library for creating isometric Minecraft renders
with Mineways and Blender.
"""

__title__            = "mcrender"
__description__      = "Command-line tool and Python library for creating isometric Minecraft renders with Mineways and Blender."
__url__              = "https://github.com/avdstaaij/mcrender"
__author__           = "Arthur van der Staaij"
__author_email__     = "arthurvanderstaaij@gmail.com"
__license__          = "GPL-3.0-or-later"
__copyright__        = "Copyright (c) 2023 Arthur van der Staaij"
__version__          = "1.0.1"


# ==================================================================================================


from typing import Optional
from dataclasses import dataclass
from functools import lru_cache
import time
import subprocess
import os
import shutil
from tempfile import TemporaryDirectory
from configparser import ConfigParser

import platformdirs
from PIL import Image

from mcrender._util import eprint, move


_SCRIPT_DIR          = os.path.dirname(os.path.realpath(__file__))
_BLENDER_BLEND_PATH  = f"{_SCRIPT_DIR}/_data/blender/mineways-isometric.blend"
_BLENDER_SCRIPT_PATH = f"{_SCRIPT_DIR}/_data/blender/mineways-isometric.py.txt"
_DEFAULT_CONFIG_PATH = f"{_SCRIPT_DIR}/_data/default-config.conf"
CONFIG_PATH          = os.path.join(platformdirs.user_config_dir("mcrender", ensure_exists=True), "config.conf")

_DIMENSION_ID_TO_MINEWAYS_NAME = {
    "overworld":  "Overworld",
    "the_nether": "Nether",
    "nether":     "Nether",
    "the_end":    "The End",
    "end":        "The End"
}
DIMENSIONS = list(_DIMENSION_ID_TO_MINEWAYS_NAME.keys())


# --------------------------------------------------------------------------------------------------
# Exception classes


class MCRenderError(Exception):
    """Base class for all mcrender exceptions."""


class ConfigAccessError(MCRenderError):
    """Raised when the config file cannot be accessed."""


class MinewaysCommandNotSetError(MCRenderError):
    """Raised when the Mineways command is not set."""


class MinewaysLaunchError(MCRenderError):
    """Raised when the Mineways command cannot be launched."""


class MinewaysError(MCRenderError):
    """Raised when Mineways returns an error."""


class MinewaysBadWorldError(MCRenderError):
    """Raised when Mineways cannot load the world."""


class BlenderCommandNotSetError(MCRenderError):
    """Raised when the Blender command is not set."""


class BlenderLaunchError(MCRenderError):
    """Raised when the Blender command cannot be launched."""


class BlenderError(MCRenderError):
    """Raised when Blender returns an error."""


class OutputFileExistsError(MCRenderError):
    """Raised when an output file or directory already exists and cannot be overwritten."""


# --------------------------------------------------------------------------------------------------
# Config file


@dataclass
class _Config:
    mineways_cmd: Optional[str] = "mineways"
    blender_cmd:  Optional[str] = "blender"


def ensure_config_file():
    """Creates a default config file if it doesn't exist.

    Raises a ConfigAccessError if creation fails.
    """

    if not os.path.isfile(CONFIG_PATH):
        try:
            shutil.copy2(_DEFAULT_CONFIG_PATH, CONFIG_PATH)
        except OSError as e:
            raise ConfigAccessError(f"Could not create default config file at {CONFIG_PATH}") from e


@lru_cache(maxsize=None)
def _read_config_file():
    """Reads and parses the config file.

    If the config file doesn't exist, creates the default one.

    Raises a ConfigAccessError if the config file cannot be accessed.
    """

    ensure_config_file()
    parser = ConfigParser()
    try:
        with open(CONFIG_PATH, "r", encoding="utf-8") as file:
            parser.read_string("[DEFAULT]\n" + file.read())
    except OSError as e:
        raise ConfigAccessError(f"Could not read config file {CONFIG_PATH}") from e

    return _Config(
        mineways_cmd = parser.get("DEFAULT", "mineways-cmd", fallback=None),
        blender_cmd  = parser.get("DEFAULT", "blender-cmd",  fallback=None),
    )


# --------------------------------------------------------------------------------------------------
# Mineways


def mineways_make_obj(
    output_dir_path: str,
    output_name:     str,
    world_path:      str,
    x:               int,
    y:               int,
    z:               int,
    size_x:          int,
    size_y:          int,
    size_z:          int,
    rotation:        int = 0,
    dimension:       str = "overworld",
    mineways_cmd:    Optional[str] = None
 ):
    """Creates an OBJ model of a Minecraft world snippet using Mineways.

    The output OBJ file is written to `<output_dir_path>/<output_name>.obj`.
    Auxillary model files (like textures) are written to `<output_dir_path>`.

    Converts the snippet specified by the Minecraft world path <world_path>, the coordinate box
    <x>, <y>, <z>, <size_x>, <size_y> and <size_z>, and the dimension id <dimension>. The sizes must
    all be strictly positive. The view direction is optionally rotated by <rotation> (0, 1, 2 or 3).

    Uses <mineways_cmd> to run Mineways. If <mineways_cmd> is None, uses the command defined in the
    config file.

    Raises:
    - `ConfigAccessError` if the config file cannot be accessed.
    - `MinewaysCommandNotSetError` if the mineways command is set neither in the config file nor as
      an argument.
    - `MinewaysLaunchError` if Mineways cannot not be lauched.
    - `MinewaysError` if Mineways returns an error.
    - `MinewaysBadWorldError` if Mineways cannot load the world.

    May raise other exceptions as well, such as `OSError`.
    """

    if size_x <= 0 or size_y <= 0 or size_z <= 0:
        raise ValueError("The size must be positive in each dimension.")

    if rotation not in (0, 1, 2, 3):
        raise ValueError("The rotation must be 0, 1, 2 or 3.")

    if dimension not in DIMENSIONS:
        raise ValueError(f"The dimension must be one of {{{', '.join(DIMENSIONS)}}}.")

    if mineways_cmd is None:
        mineways_cmd = _read_config_file().mineways_cmd
        if mineways_cmd is None:
            raise MinewaysCommandNotSetError("The Mineways command is set neither in the config file nor as an argument.")

    with TemporaryDirectory() as tmpDir:
        scriptPath = f"{tmpDir}/script.mwscript"
        logPath    = f"{tmpDir}/log.txt"

        with open(scriptPath, "w", encoding="utf-8") as f:
            # https://www.realtimerendering.com/erich/minecraft/public/mineways/scripting.html
            f.write(f"Save Log file: {tmpDir}/log.txt\n")
            f.write(f"Minecraft world: {world_path}\n")
            f.write(f"View {_DIMENSION_ID_TO_MINEWAYS_NAME[dimension]}\n")
            f.write(f"Selection location min to max: {x}, {y}, {z} to {x + size_x - 1}, {y + size_y - 1}, {z + size_z - 1}\n")
            f.write("Set render type: Wavefront OBJ absolute indices\n")
            f.write("File type: Export all textures to three large images\n")
            f.write(f"Rotate model {rotation*90} degrees\n")
            f.write("Scale model by making each block 100 cm high\n")
            f.write("Tree leaves solid: yes\n")
            f.write("Use biomes: yes\n")
            f.write(f"Export for Rendering: {output_dir_path}/{output_name}.obj\n")
            f.write("Close\n")

        os.makedirs(output_dir_path, exist_ok=True)

        # If Mineways cannot find the world, it will write an error message to the log file, but it
        # won't stop. There seems to be no command-line option to make it stop in this case, either.
        # To handle this, we run Mineways in the background, periodically check the log file, and
        # kill the process if we see the error message.

        try:
            cmd = [mineways_cmd, "-m", "-suppress", "-s", "none", scriptPath]
            process = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        except OSError as e:
            raise MinewaysLaunchError(f"Mineways could not be launched: {e}") from e

        while True:
            if process.poll() is not None:
                if process.returncode != 0:
                    raise MinewaysError(f"Mineways returned an error ({process.returncode})")
                break
            try:
                with open(logPath, "r", encoding="utf-8") as f:
                    for line in f:
                        if line.startswith("Error reading line 2: Mineways attempted to load world"):
                            process.kill()
                            raise MinewaysBadWorldError(f"Mineways could not load the world {repr(world_path)}")
            except FileNotFoundError:
                pass
            time.sleep(0.1)


# --------------------------------------------------------------------------------------------------
# Blender


def blender_render_obj(
    output_path: str,
    obj_path:    str,
    exposure:    float = 0,
    trim:        bool  = True,
    force:       bool  = False,
    blender_cmd: Optional[str] = None
):
    """Renders an OBJ file (as created by Mineways) to a PNG image using Blender.

    The output PNG file is written to <output_path>.

    The post-processing exposure can optionally be modified by <exposure> (can be negative).

    The image is created at a resolution of 2048x2048 pixels, but the model may not be square.
    If <trim> is True, the image is trimmed down to the model's bounding box.
    Otherwise, the model will be centered in the image.

    If <force> is true, overwrites any existing file at <output_path>.

    Uses <blender_cmd> to run Blender. If <blender_cmd> is None, uses the command defined in the
    config file.

    Raises:
    - `ConfigAccessError` if the config file cannot be accessed.
    - `BlenderCommandNotSetError` if the Blender command is set neither in the config file nor as an
      argument.
    - `BlenderLaunchError` if Blender cannot be launched.
    - `BlenderError` if Blender returns an error.
    - `OutputFileExistsError` if the output file already exists and <force> is false.

    May raise other exceptions as well, such as `OSError`.
    """

    if blender_cmd is None:
        blender_cmd = _read_config_file().blender_cmd
        if blender_cmd is None:
            raise BlenderCommandNotSetError("The Blender command is set neither in the config file nor as an argument.")

    with TemporaryDirectory() as tmpDir:
        try:
            cmd  = [blender_cmd, "--background", _BLENDER_BLEND_PATH, "--python", _BLENDER_SCRIPT_PATH, "--", "--exposure", str(exposure), obj_path, f"{tmpDir}/output"]
            subprocess.run(cmd, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        except OSError as e:
            raise BlenderLaunchError(f"Blender could not be launched: {e}") from e
        except subprocess.CalledProcessError as e:
            raise BlenderError(f"Blender returned an error: {e}") from e

        if trim:
            with Image.open(f"{tmpDir}/output0001.png") as image:
                trimmed = image.crop(image.getbbox())
            trimmed.save(f"{tmpDir}/output0001.png")

        if os.path.dirname(output_path): os.makedirs(os.path.dirname(output_path), exist_ok=True)
        try:
            move(f"{tmpDir}/output0001.png", output_path, overwrite=force, never_overwrite_dir=True)
        except FileExistsError as e:
            raise OutputFileExistsError(str(e)) from e


# --------------------------------------------------------------------------------------------------
# Mineways + Blender


def render(
    output_path:  str,
    world_path:   str,
    x:            int,
    y:            int,
    z:            int,
    size_x:       int,
    size_y:       int,
    size_z:       int,
    rotation:     int   = 0,
    dimension:    str   = "overworld",
    exposure:     float = 0,
    trim:         bool  = True,
    force:        bool  = False,
    mineways_cmd: Optional[str] = None,
    blender_cmd:  Optional[str] = None,
    verbose:      bool  = False,
):
    """Renders a Minecraft world snippet to a PNG image using Mineways and Blender.

    The output PNG file is written to <output_path>.

    Renders the snippet specified by the Minecraft world path <world_path>, the coordinate box
    <x>, <y>, <z>, <size_x>, <size_y> and <size_z>, and the dimension id <dimension>. The sizes must
    all be strictly positive. The view direction is optionally rotated by <rotation> (0, 1, 2 or 3).

    The post-processing exposure can optionally be modified by <exposure> (can be negative).

    The image is created at a resolution of 2048x2048 pixels, but the model may not be square.
    If <trim> is True, the image is trimmed down to the model's bounding box.
    Otherwise, the model will be centered in the image.

    If <force> is true, overwrites any existing file at <output_path>.

    Uses <mineways_cmd> to run Mineways and <blender_cmd> to run Blender.
    If either argument is None, uses the command defined in the config file.

    Raises:
    - `ConfigAccessError` if the config file cannot be accessed.
    - `MinewaysCommandNotSetError` if the mineways command is set neither in the config file nor as
      an argument.
    - `MinewaysLaunchError` if Mineways cannot not be lauched.
    - `MinewaysError` if Mineways returns an error.
    - `MinewaysBadWorldError` if Mineways cannot load the world.
    - `BlenderCommandNotSetError` if the Blender command is set neither in the config file nor as an
      argument.
    - `BlenderLaunchError` if Blender cannot be launched.
    - `BlenderError` if Blender returns an error.
    - `OutputFileExistsError` if the output file already exists and <force> is false.

    May raise other exceptions as well, such as `OSError`.
    """
    with TemporaryDirectory() as tmpDir:
        if verbose: eprint("Running Mineways...")
        mineways_make_obj(tmpDir, "snippet", world_path, x, y, z, size_x, size_y, size_z, rotation, dimension, mineways_cmd)
        if verbose: eprint("Rendering...")
        blender_render_obj(output_path, f"{tmpDir}/snippet.obj", exposure, trim, force, blender_cmd)
        if verbose: eprint(f"Created {output_path}")



"""mcrender command-line interface."""

from typing import Optional, Tuple
import sys

from click import UsageError
import cloup

import mcrender
from mcrender._util import eprint


def parse_box_spec(pos: Tuple[Tuple[int]], size: Optional[Tuple[int]]) -> Tuple[int]:
    """Parses specified pos and size arguments into a (x,y,z,sx,sy,sz) tuple."""

    def raise_box_spec_error():
        raise UsageError(
            "The box to render is not specified.\n"
            "You can specify the box to render in two ways:\n"
            "1. Using --pos and --size.\n"
            "2. Using two --pos options: one for each corner (inclusive)."
        )

    if len(pos) == 1:
        if size is None:
            raise_box_spec_error()
        return *pos[0], *size

    if len(pos) == 2:
        if size is not None:
            raise_box_spec_error()

        x1, x2 = sorted((pos[0][0], pos[1][0]))
        y1, y2 = sorted((pos[0][1], pos[1][1]))
        z1, z2 = sorted((pos[0][2], pos[1][2]))

        return x1, y1, z1, x2 - x1 + 1, y2 - y1 + 1, z2 - z1 + 1

    raise_box_spec_error()


# We have to take the position as an option instead of a positional argument,
# because otherwise you need to use "--" to pass negative numbers, which is
# quite unintuitive. For consistency, we take the size as an option as well.

epilog = (
    f"*: Dimension options: {{{', '.join(mcrender.DIMENSIONS)}}}\n"
    "\n"
    "**: The render is always created at a resolution of 2048x2048 pixels, but the model may not be square. If `--trim` is set (which it is by default), the image is trimmed down to the model's bounding box. Otherwise, the model will be centered in the image.\n"
    "\n"
    f"Config file location: {mcrender.CONFIG_PATH}\n"
)

# pylint: disable=no-value-for-parameter
@cloup.command(context_settings={"show_default": True}, epilog=epilog)
@cloup.argument("world-path",  metavar="<world path>",  type=cloup.Path(exists=True, file_okay=False))
@cloup.argument("output-path", metavar="<output path>", type=cloup.Path())
@cloup.option("--pos",  "-p",     metavar=" <x> <y> <z>", help="Render-box corner",                           type=int, nargs=3, multiple=True)
@cloup.option("--size", "-s",     metavar="<x> <y> <z>",  help="Render-box size",                             type=int, nargs=3)
@cloup.option("--rotation",       metavar="{0,1,2,3}",    help="Rotation of the camera.",                     type=int, default=0)
@cloup.option("--dimension",      metavar="<id>",         help="World dimension.*",                           type=str, default="overworld")
@cloup.option("--exposure",       metavar="<float>",      help="Exposure for post-processing.",               type=float, default=0)
@cloup.option("--trim/--no-trim",                         help="Trim the output image.**",                    default=True)
@cloup.option("--force", "-f",                            help="Overwrite any existing file at output path.", is_flag=True)
@cloup.option("--mineways-cmd",   metavar="<cmd>",        help="Command to run Mineways.",                    type=str)
@cloup.option("--blender-cmd",    metavar="<cmd>",        help="Command to run Blender.",                     type=str)
@cloup.option("--verbose", "-v", "verbose",               help="Print more information.",                     flag_value=True)
@cloup.option("--quiet",   "-q", "verbose",               help="Cancel a previous --verbose.",                flag_value=False, default=False, show_default=False)
@cloup.version_option(mcrender.__version__, package_name="mcrender", prog_name="mcrender", message="%(prog)s %(version)s")
def cli(world_path: str, output_path: str, pos: Tuple[Tuple[int]], size: Optional[Tuple[int]], rotation: int, dimension: str, exposure: float, trim: bool, force: bool, mineways_cmd: Optional[str], blender_cmd: Optional[str], verbose: bool):
    """
    Render a Minecraft world snippet with Mineways and Blender.

    \b
    You can specify the box to render in two ways:
    1. Using --pos and --size.
    2. Using two --pos options: one for each corner (inclusive).
    """

    error_prefix = "\nError: " if verbose else "Error: "

    box = parse_box_spec(pos, size)

    try:
        mcrender.render(
            output_path  = output_path,
            world_path   = world_path,
            x            = box[0],
            y            = box[1],
            z            = box[2],
            size_x       = box[3],
            size_y       = box[4],
            size_z       = box[5],
            rotation     = rotation,
            dimension    = dimension,
            exposure     = exposure,
            trim         = trim,
            force        = force,
            mineways_cmd = mineways_cmd,
            blender_cmd  = blender_cmd,
            verbose      = verbose
        )
        return
    except mcrender.ConfigAccessError as e:
        eprint(f"{error_prefix}{e}\n    {e.__cause__}", file=sys.stderr)
    except mcrender.MinewaysCommandNotSetError:
        eprint(f"{error_prefix}You must either set --mineways-cmd, or set mineways-cmd in\n{repr(mcrender.CONFIG_PATH)}.")
    except mcrender.BlenderCommandNotSetError:
        eprint(f"{error_prefix}You must either set --blender-cmd, or set blender-cmd in\n{repr(mcrender.CONFIG_PATH)}.")
    except mcrender.MinewaysLaunchError as e:
        eprint(f"{error_prefix}Mineways could not be launched.\n    {e.__cause__}\n")
        if mineways_cmd is not None:
            eprint(
                 "Make sure that you have downloaded Mineways and that your specified",
                f"command ({repr(mineways_cmd)}) runs it.",
                sep="\n"
            )
        else:
            eprint(
                "Make sure that you have downloaded Mineways and that your configured",
                f"mineways-cmd (in {repr(mcrender.CONFIG_PATH)}) runs it.",
                sep="\n"
            )
    except mcrender.BlenderLaunchError as e:
        eprint(f"{error_prefix}Blender could not be launched.\n    {e.__cause__}\n")
        if blender_cmd is not None:
            eprint(
                 "Make sure that you have downloaded Blender and that your specified",
                f"command ({repr(blender_cmd)}) runs it.",
                sep="\n"
            )
        else:
            eprint(
                "Make sure that you have downloaded Blender and that your configured",
                f"blender-cmd (in {repr(mcrender.CONFIG_PATH)}) runs it.",
                sep="\n"
            )
    except mcrender.MinewaysError as e:
        eprint(f"{error_prefix}{e}")
    except mcrender.MinewaysBadWorldError as e:
        eprint(
            f"{error_prefix}Mineways could not load the specified world",
            f"{repr(world_path)}.",
            "",
            "There might still be a Mineways window open that you'll have to close",
            "manually. Sorry about that, I'm afraid I don't know how to prevent it.",
            sep="\n"
        )
    except mcrender.BlenderError as e:
        eprint(f"{error_prefix}Blender returned an error.\n    {e.__cause__}")
    except mcrender.OutputFileExistsError as e:
        eprint(f"{error_prefix}{e}")

    sys.exit(1)


def main():
    try:
        mcrender.ensure_config_file()
    except mcrender.ConfigAccessError as e:
        eprint(f"Warning: {e}\n    {e.__cause__}\n")
        eprint("If this execution requires the config file, it will fail.\n")

    cli()



import os
from setuptools import setup


SCRIPT_DIR         = os.path.abspath(os.path.dirname(__file__))
METADATA_FILE_PATH = os.path.join(SCRIPT_DIR, "mcrender/__init__.py")


# Based on https://github.com/pypa/pip/blob/9aa422da16e11b8e56d3597f34551f983ba9fbfd/setup.py
with open(METADATA_FILE_PATH) as file:
    metadata_file_lines = file.read().splitlines()

def get_metadata(name: str) -> str:
    for line in metadata_file_lines:
        dunderString = f"__{name}__"
        if line.startswith(f"{dunderString}"):
            # __{name}__ = "{value}"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError(f"Unable to find {dunderString} value.")


with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()


setup(
    name                          = get_metadata("title"),
    version                       = get_metadata("version"),
    description                   = get_metadata("description"),
    long_description              = long_description,
    long_description_content_type = "text/markdown",
    url                           = get_metadata("url"),
    author                        = get_metadata("author"),
    author_email                  = get_metadata("author_email"),
    license                       = get_metadata("license"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Artistic Software",
        "Topic :: Games/Entertainment :: Simulation",
        "Topic :: Multimedia :: Graphics :: 3D Modeling",
        "Topic :: Multimedia :: Graphics :: 3D Rendering",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Version Control :: Git",
    ],
    keywords="Minecraft, Mineways, Blender, rendering, isometric, 3D, 3D modeling, 3D rendering, cli, automation",
    project_urls={
        "Bug Reports": "https://github.com/avdstaaij/mcrender/issues",
        "Source":      "https://github.com/avdstaaij/mcrender",
    },
    packages = ["mcrender"], # Note: subpackages must be listed explicitly
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "mcrender = mcrender._cli:main",
        ],
    },
    python_requires=">=3.7, <4",
    install_requires=[
        "cloup >= 3.0.0",
        "Pillow",
        "platformdirs >= 3.2.0",
    ],
    zip_safe=False
)
